// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_frm_wt_async_fifo_core.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE			: 08/22/2013
// =============================================================================
// DESCRIPTION	: async fifo controller
//
// =============================================================================
// REVISION HISTORY	:
//	v1.0	: Initial version release
//
// =============================================================================
// PARAMETERS	:
//		FIFO_DEPTH		: FIFO depth
//		DATA_BIT_WIDTH	: data bit width
//		WATERMARK_W		: watermark of asserting almost full flag
//						: ( word count of remaining writable space )
//		WATERMARK_R		: watermark of asserting almost empty flag
//						: ( word count of remaining readable space )
//		RETIME			: read data retiming setting 0:no retiming / 1:retiming
//
// =============================================================================
// I/O PORTS	:
//		wt_clk			: write clock
//		rd_clk			: read clock
//		rst_n			: asynchronous reset
//		srst			: synchronous reset
//	
//		wt_en_in		: write enable
//		rd_en_in		: read enable
//		data_in 		: write data
//		data_out		: read data
//		a_empty_out 	: almost empty flag
//		empty_out		: empty flag
//		a_full_out		: almost full flag
//		full_out		: full flag
//	
//		ram_wt_en_out	: ram write enable
//		ram_rd_en_out	: ram read enable
//		ram_wt_adr_out	: ram write adress
//		ram_rd_adr_out	: ram read adress
//		ram_wt_data_out : ram write data
//		ram_rd_data_in	: ram read data
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module avb_frm_wt_async_fifo_core	(
	wt_clk			,	//	write clock
	rd_clk			,	//	read clock
	rst_n			,	//	asynchronous reset
	srst			,	//	synchronous reset

	wt_en_in		,	//	write enable
	rd_en_in		,	//	read enable
	data_in 		,	//	write data
	data_out		,	//	read data
	a_empty_out 	,	//	almost empty flag
	empty_out		,	//	empty flag
	a_full_out		,	//	almost full flag
	full_out		,	//	full flag

	ram_wt_en_out	,	//	ram write enable
	ram_rd_en_out	,	//	ram read enable
	ram_wt_adr_out	,	//	ram write adress
	ram_rd_adr_out	,	//	ram read adress
	ram_wt_data_out ,	//	ram write data
	ram_rd_data_in		//	ram read data
) ;

// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter	FIFO_DEPTH				= 8 	;
	parameter	DATA_BIT_WIDTH			= 16	;
	parameter	WATERMARK_W 			= 1 	;
	parameter	WATERMARK_R 			= 1 	;
	parameter	RETIME					= 0 	;

// =============================================================================
// PORT DECLARATION
// =============================================================================
	input	wire								wt_clk			;	//	write clock
	input	wire								rd_clk			;	//	read clock
	input	wire								rst_n			;	//	asynchronous reset
	input	wire								srst			;	//	synchronous reset

	input	wire								wt_en_in		;	//	write enable
	input	wire								rd_en_in		;	//	read enable
	input	wire	[ DATA_BIT_WIDTH-1 : 0 ]	data_in 		;	//	write data
	output	wire	[ DATA_BIT_WIDTH-1 : 0 ]	data_out		;	//	read data
	output	wire								a_empty_out 	;	//	almost empty flag
	output	wire								empty_out		;	//	empty flag
	output	wire								a_full_out		;	//	almost full flag
	output	wire								full_out		;	//	full flag

	output	wire								ram_wt_en_out	;	//	ram write enable
	output	wire								ram_rd_en_out	;	//	ram read enable
	output	wire	[	  FIFO_DEPTH-1 : 0 ]	ram_wt_adr_out	;	//	ram write adress
	output	wire	[	  FIFO_DEPTH-1 : 0 ]	ram_rd_adr_out	;	//	ram read adress
	output	wire	[ DATA_BIT_WIDTH-1 : 0 ]	ram_wt_data_out ;	//	ram write data
	input	wire	[ DATA_BIT_WIDTH-1 : 0 ]	ram_rd_data_in	;	//	ram read data

// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================
	reg 	[	  FIFO_DEPTH-1 : 0 ]	wrp_gray_ff 			;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	wrp_gray_a_ff			;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	rdp_sync_ff 			;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	rdp_w_ff				;
	wire	[	  FIFO_DEPTH-1 : 0 ]	sub_pointer 			;
	wire	[	  FIFO_DEPTH-1 : 0 ]	sub_afr_pointer 		;
	reg 								afull_ff				;
	reg 								full_ff 				;

	reg 	[	  FIFO_DEPTH-1 : 0 ]	wrp_sync_ff 			;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	wrp_r_ff				;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	rdp_gray_ff 			;
	reg 	[	  FIFO_DEPTH-1 : 0 ]	rdp_gray_a_ff			;
	reg 								empty_ff				;
	reg 								aempty_ff				;
	wire	[	  FIFO_DEPTH-1 : 0 ]	sub_r_pointer			;
	wire	[	  FIFO_DEPTH-1 : 0 ]	sub_aer_pointer 		;

//	wire	[	  FIFO_DEPTH-1 : 0 ]	wrp_g2b 				;
//	wire	[	  FIFO_DEPTH-1 : 0 ]	wrp_b					;
	reg		[     FIFO_DEPTH-1 : 0 ]	wrp_b_ff				; //v1.1
	wire	[     FIFO_DEPTH-1 : 0 ]	wrp_b_nxt				; //v1.1
	wire	[	  FIFO_DEPTH-1 : 0 ]	wrp_gray				;

//	wire	[	  FIFO_DEPTH-1 : 0 ]	rdp_g2b 				;
//	wire	[	  FIFO_DEPTH-1 : 0 ]	rdp_b					;
	reg		[     FIFO_DEPTH-1 : 0 ]    rdp_b_ff				; //v1.1
	wire	[     FIFO_DEPTH-1 : 0 ]	rdp_b_nxt				; //v1.1
	wire	[	  FIFO_DEPTH-1 : 0 ]	rdp_gray				;

	wire	[	  FIFO_DEPTH-1 : 0 ]	rdp_w_g2b				;
	wire	[	  FIFO_DEPTH-1 : 0 ]	wrp_r_g2b				;

//	reg 	[	  FIFO_DEPTH-1 : 0 ]	wradr_ff				;
//	reg 	[	  FIFO_DEPTH-1 : 0 ]	rdadr_ff				;
	reg 	[ DATA_BIT_WIDTH-1 : 0 ]	data_out_retime0_ff 	;
	wire								retimemode				;

	wire								set_wrp_gray_v			;
	wire								set_rdp_gray_v			;

	reg 								srst_wclk_lat1_ff		;
	reg 								srst_wclk_lat2_ff		;
	reg 								wt_srst_ff				;
	reg 								wt_srst_lat_ff			;
	reg 								wt_srst_rclk_lat1_ff	;
	reg 								wt_srst_rclk_lat2_ff	;
	reg 								rd_srst_clr_ff			;
	reg 								srst_rclk_lat1_ff		;
	reg 								srst_rclk_lat2_ff		;
	reg 								rd_srst_ff				;
	reg 								rd_srst_lat_ff			;
	reg 								rd_srst_wclk_lat1_ff	;
	reg 								rd_srst_wclk_lat2_ff	;
	reg 								wt_srst_clr_ff			;

// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================
//---------------------//
//--  Output Signal  --//
//---------------------//
	assign	a_empty_out 	= aempty_ff ;
	assign	empty_out		= empty_ff ;
	assign	a_full_out		= afull_ff ;
	assign	full_out		= full_ff ;
	assign	data_out		= ( RETIME ) ?	data_out_retime0_ff : ram_rd_data_in ;

	assign	ram_wt_adr_out	=  wrp_b_ff ;
	assign	ram_rd_adr_out	=  rdp_b_ff ;
//	assign	ram_wt_adr_out	= wradr_ff ;
//	assign	ram_rd_adr_out	= rdadr_ff ;
	assign	ram_wt_en_out	= set_wrp_gray_v ;
	assign	ram_rd_en_out	= set_rdp_gray_v ;
	assign	ram_wt_data_out = data_in ;

	assign	set_wrp_gray_v	= wt_en_in & ( ~ full_ff ) ;
	assign	set_rdp_gray_v	= rd_en_in & ( ~ empty_ff ) ;

//-----------------------------------------//
//--  Write Pointer (Gray Code Counter)  --//
//-----------------------------------------//
	assign	wrp_b_nxt	= wrp_b_ff + 1'b1 ; //v1.1
	assign	wrp_gray	= ( { 1'b0 , wrp_b_nxt[FIFO_DEPTH-1:1] } ) ^ wrp_b_nxt ;
//	assign wrp_g2b	= ( { 1'b0 , wrp_g2b[ FIFO_DEPTH-1 : 1 ] } ) ^ wrp_gray_ff;
//	assign wrp_b	= wrp_g2b + 1'b1;
//	assign wrp_gray = ( { 1'b0 , wrp_b[ FIFO_DEPTH-1 : 1 ] } ) ^ wrp_b ;

	always @( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wrp_gray_ff 	<= { FIFO_DEPTH{1'b0} } ;
			wrp_gray_a_ff	<= { FIFO_DEPTH{1'b0} } ;
			wrp_b_ff		<= { FIFO_DEPTH{1'b0} } ; //v1.1
		end
		else if ( wt_srst_ff ) begin
			wrp_gray_ff 	<= { FIFO_DEPTH{1'b0} } ;
			wrp_gray_a_ff	<= { FIFO_DEPTH{1'b0} } ;
			wrp_b_ff		<= { FIFO_DEPTH{1'b0} } ; //v1.1
		end
		else if ( set_wrp_gray_v ) begin
			wrp_gray_ff 	<= wrp_gray ;
			wrp_gray_a_ff	<= wrp_gray ;
			wrp_b_ff		<= wrp_b_nxt ; //v1.1
		end
	end

//----------------------//
//--  clk_w -> clk_r  --//
//----------------------//
	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wrp_sync_ff	<= { FIFO_DEPTH{1'b0} } ;
			wrp_r_ff	<= { FIFO_DEPTH{1'b0} } ;
		end
//		else if ( rd_srst_ff ) begin
//			wrp_sync_ff	<= { FIFO_DEPTH{1'b0} } ;
//			wrp_r_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
		else begin
			wrp_sync_ff	<= wrp_gray_a_ff ;
			wrp_r_ff	<= wrp_sync_ff ;
		end
	end

//----------------------------------------//
//--  Read Pointer (Gray Code Counter)	--//
//----------------------------------------//
	assign	rdp_b_nxt		= rdp_b_ff + 1'b1 ; //v1.1
	assign	rdp_gray	= ( { 1'b0 , rdp_b_nxt[FIFO_DEPTH-1:1] } ) ^ rdp_b_nxt ;
//	assign rdp_g2b	= ( { 1'b0 , rdp_g2b[ FIFO_DEPTH-1 : 1 ] } ) ^ rdp_gray_ff ;
//	assign rdp_b	= rdp_g2b + 1'b1;
//	assign rdp_gray = ( { 1'b0 , rdp_b[ FIFO_DEPTH-1 : 1 ] } ) ^ rdp_b ;

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rdp_gray_ff 	<= { FIFO_DEPTH{1'b0} } ;
			rdp_gray_a_ff	<= { FIFO_DEPTH{1'b0} } ;
			rdp_b_ff		<= { FIFO_DEPTH{1'b0} } ; //v1.1
		end
		else if ( rd_srst_ff ) begin
			rdp_gray_ff 	<= { FIFO_DEPTH{1'b0} } ;
			rdp_gray_a_ff	<= { FIFO_DEPTH{1'b0} } ;
			rdp_b_ff		<= { FIFO_DEPTH{1'b0} } ; //v1.1
		end
		else if (set_rdp_gray_v) begin
			rdp_gray_ff 	<= rdp_gray ;
			rdp_gray_a_ff	<= rdp_gray ;
			rdp_b_ff		<= rdp_b_nxt ; //v1.1
		end
	end

//----------------------------------------//
//--  almost empty and empty checking.	--//
//----------------------------------------//
	assign	wrp_r_g2b		= { 1'b0 , wrp_r_g2b[ FIFO_DEPTH-1 : 1 ] } ^ wrp_r_ff ;
	assign	sub_r_pointer	= wrp_r_g2b - rdp_b_nxt ; //v1.1
	assign	sub_aer_pointer	= wrp_r_g2b - rdp_b_ff ; //v1.1
//	assign	sub_r_pointer	= wrp_r_g2b - rdp_b ;
//	assign	sub_aer_pointer = wrp_r_g2b - rdp_g2b ;

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			aempty_ff	<= 1'b1 ;
		end
		else if ( rd_srst_ff ) begin
			aempty_ff	<= 1'b1 ;
		end
		else if ( set_rdp_gray_v && ( sub_r_pointer <= WATERMARK_R ) ) begin
			aempty_ff	<= 1'b1 ;
		end
		else if (sub_aer_pointer > WATERMARK_R) begin
			aempty_ff	<= 1'b0 ;
		end
	end

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			empty_ff	<= 1'b1 ;
		end
		else if ( rd_srst_ff ) begin
			empty_ff	<= 1'b1 ;
		end
		else if ( set_rdp_gray_v && ( rdp_gray == wrp_r_ff ) ) begin
			empty_ff	<= 1'b1 ;
		end
		else if ( rdp_gray_ff != wrp_r_ff ) begin
			empty_ff	<= 1'b0 ;
		end
	end

//----------------------//
//--  clk_r -> clk_w  --//
//----------------------//
	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rdp_sync_ff	<= { FIFO_DEPTH{1'b0} } ;
			rdp_w_ff	<= { FIFO_DEPTH{1'b0} } ;
		end
//		else if ( wt_srst_ff ) begin
//			rdp_sync_ff	<= { FIFO_DEPTH{1'b0} } ;
//			rdp_w_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
		else begin
			rdp_sync_ff	<= rdp_gray_a_ff ;
			rdp_w_ff	<= rdp_sync_ff ;
		end
	end

//--------------------------------------//
//--  almost full and full checking.  --//
//--------------------------------------//
	assign rdp_w_g2b		= ( { 1'b0 , rdp_w_g2b[ FIFO_DEPTH-1 : 1 ] } ) ^ rdp_w_ff ;
	assign sub_pointer		= rdp_w_g2b - wrp_b_nxt ; //v1.1
	assign sub_afr_pointer	= rdp_w_g2b - wrp_b_ff ;  //v1.1
//	assign sub_pointer		= rdp_w_g2b - wrp_b ;
//	assign sub_afr_pointer	= rdp_w_g2b - wrp_g2b ;

	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			afull_ff	<= 1'b0 ;
		end
		else if ( wt_srst_ff || srst_wclk_lat2_ff ) begin
			afull_ff	<= 1'b1 ;
		end
		else if ( wt_srst_clr_ff ) begin
			afull_ff	<= 1'b0 ;
		end
		else if ( set_wrp_gray_v && (sub_pointer <= WATERMARK_W ) ) begin
			afull_ff	<= 1'b1 ;
		end
		else if (sub_afr_pointer > WATERMARK_W ) begin
			afull_ff	<= 1'b0 ;
		end
	end

	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			full_ff	<= 1'b0 ;
		end
		else if ( wt_srst_ff || srst_wclk_lat2_ff ) begin
			full_ff	<= 1'b1 ;
		end
		else if ( wt_srst_clr_ff ) begin
			full_ff	<= 1'b0 ;
		end
		else if ( set_wrp_gray_v && ( rdp_w_ff == wrp_gray ) ) begin
			full_ff	<= 1'b1 ;
		end
		else if ( rdp_w_ff != wrp_gray_ff ) begin
			full_ff	<= 1'b0 ;
		end
	end

//--------------------//
//--  FiFo Control	--//
//--------------------//
//	always @ ( posedge wt_clk or negedge rst_n ) begin
//		if ( !rst_n ) begin
//			wradr_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
//		else if ( wt_srst_ff ) begin
//			wradr_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
//		else if ( set_wrp_gray_v ) begin
//			wradr_ff	<= wradr_ff + 1'b1 ;
//		end
//	end
//
//	always @ ( posedge rd_clk or negedge rst_n ) begin
//		if ( !rst_n ) begin
//			rdadr_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
//		else if ( rd_srst_ff ) begin
//			rdadr_ff	<= { FIFO_DEPTH{1'b0} } ;
//		end
//		else if ( set_rdp_gray_v ) begin
//			rdadr_ff	<= rdadr_ff + 1'b1 ;
//		end
//	end

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			data_out_retime0_ff	<= { DATA_BIT_WIDTH{1'b0} } ;
		end
		else if ( rd_srst_ff ) begin
			data_out_retime0_ff	<= { DATA_BIT_WIDTH{1'b0} } ;
		end
		else if ( set_rdp_gray_v ) begin
			data_out_retime0_ff	<= ram_rd_data_in ;
		end
	end

//-------------------------//
//--  synchronous reset  --//
//-------------------------//
//	write clock srst
	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wt_srst_ff	<= 1'b0 ;
		end
		else if ( srst_wclk_lat2_ff ) begin
			wt_srst_ff	<= 1'b1 ;
		end
		else if ( wt_srst_clr_ff && !srst_wclk_lat2_ff ) begin
			wt_srst_ff	<= 1'b0 ;
		end
	end

	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			srst_wclk_lat1_ff	<= 1'b0 ;
			srst_wclk_lat2_ff	<= 1'b0 ;
			wt_srst_lat_ff		<= 1'b0 ;
		end
		else begin
			srst_wclk_lat1_ff	<= srst ;
			srst_wclk_lat2_ff	<= srst_wclk_lat1_ff ;
			wt_srst_lat_ff		<= wt_srst_ff ;
		end
	end

	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rd_srst_wclk_lat1_ff	<= 1'b0 ;
			rd_srst_wclk_lat2_ff	<= 1'b0 ;
		end
		else begin
			rd_srst_wclk_lat1_ff	<= rd_srst_lat_ff ;
			rd_srst_wclk_lat2_ff	<= rd_srst_wclk_lat1_ff ;
		end
	end

	always @ ( posedge wt_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wt_srst_clr_ff	<= 1'b0 ;
		end
		else if ( !wt_srst_ff ) begin
			wt_srst_clr_ff	<= 1'b0 ;
		end
		else if ( rd_srst_wclk_lat2_ff ) begin
			wt_srst_clr_ff	<= 1'b1 ;
		end 
	end

//	read clock srst
	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rd_srst_ff	<= 1'b0 ;
		end
		else if ( srst_rclk_lat2_ff ) begin
			rd_srst_ff	<= 1'b1 ;
		end
		else if ( rd_srst_clr_ff && !srst_rclk_lat2_ff ) begin
			rd_srst_ff	<= 1'b0 ;
		end
	end

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			srst_rclk_lat1_ff	<= 1'b0 ;
			srst_rclk_lat2_ff	<= 1'b0 ;
			rd_srst_lat_ff		<= 1'b0 ;
		end
		else begin
			srst_rclk_lat1_ff	<= srst ;
			srst_rclk_lat2_ff	<= srst_rclk_lat1_ff ;
			rd_srst_lat_ff		<= rd_srst_ff ;
		end
	end

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			wt_srst_rclk_lat1_ff	<= 1'b0 ;
			wt_srst_rclk_lat2_ff	<= 1'b0 ;
		end
		else begin
			wt_srst_rclk_lat1_ff	<= wt_srst_lat_ff ;
			wt_srst_rclk_lat2_ff	<= wt_srst_rclk_lat1_ff ;
		end
	end

	always @ ( posedge rd_clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			rd_srst_clr_ff	<= 1'b0 ;
		end
		else if ( !rd_srst_ff ) begin
			rd_srst_clr_ff	<= 1'b0 ;
		end
		else if ( wt_srst_rclk_lat2_ff ) begin
			rd_srst_clr_ff	<= 1'b1 ;
		end 
	end


endmodule

`default_nettype wire
