// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_frm_wt_ram.v
// DEPARTMENT	: MACNICA Advanced Technology Div.
// AUTHOR		: Yuichi Shinoda
// DATE 		: 08/03/2012
// =============================================================================
// DESCRIPTION	: Pseudo dual port ram with single write port & single read port 
//
// =============================================================================
// REVISION HISTORY	:
//	v1.0	: Initial version release
//
// =============================================================================
// PARAMETERS	:
//		ADR_BIT		: Address bit width
//		DATA_BIT	: Data bit width
//		DATA_DEPTH	: Data depth
//
// =============================================================================
// I/O PORTS		:
//		wclk		: clock for write port
//		rclk		: clock for read port
//		
//		wadr_i		: write address
//		radr_i		: read address
//		wen_i		: write enable
//		ren_i		: read enable
//		
//		wdata_i		: write data
//		rdata_o		: read data
//
// =============================================================================
`default_nettype none
`timescale 1ps / 1ps

module avb_frm_wt_ram (
	wclk	,
	rclk	,
	
	wadr_i	,
	radr_i	,
	wen_i	,
	ren_i	,
	
	wdata_i	,
	rdata_o	
) ;
	
// =============================================================================
// PARAMETER DEFINITION
// =============================================================================
	
	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	
	parameter	ADR_BIT				= 16 ;
	parameter	DATA_BIT			= 8 ;
	parameter	DATA_DEPTH			= 4096 ;
	
	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	
	
	
// =============================================================================
// PORT DECLARATION
// =============================================================================
	
	input	wire							wclk					;
	input	wire							rclk					;
	
	input	wire	[  ADR_BIT-1 : 0 ]		wadr_i					;
	input	wire	[  ADR_BIT-1 : 0 ]		radr_i					;
	input	wire							wen_i					;
	input	wire							ren_i					;
	
	input	wire	[ DATA_BIT-1 : 0 ]		wdata_i					;
	output	wire	[ DATA_BIT-1 : 0 ]		rdata_o					;
	
	
	
// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================
	
	reg		[ DATA_BIT-1 : 0 ]		mem_ff	[ DATA_DEPTH-1 : 0 ]	;
	reg		[  ADR_BIT-1 : 0 ]		radr_ff							;
	
	
	
// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================
	
	always @( posedge wclk ) begin
		if ( wen_i ) begin
			mem_ff[wadr_i]	<= wdata_i ;
		end
	end
	
	always @( posedge rclk ) begin
		if ( ren_i ) begin
			radr_ff	<= radr_i ;
		end
	end
	
	assign rdata_o	= mem_ff[radr_ff] ;
	
	
	
endmodule
`default_nettype wire
