// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_frm_wt_reg_ff.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Youhei Nagano
// DATE			: 08/22/2013
// =============================================================================
// DESCRIPTION	:Registers
//
// =============================================================================
// REVISION HISTORY	:
//	v1.0	: Initial version release
//
// =============================================================================
// PARAMETERS	:
//		ADW		: address bit width
//		DTW		: data bit width
//		ADR		: address
//		INI		: initial data
//		VLD		: generate register on/off every bit
//		PLS		: generate pulse output on/off every bit
//		TYP		: register type
//				: 0: read/write register
//				: 1: write-clear register
//				: 2: read-only register
//				: 3: read-clear register
//		
//		WPO		: write pulse width ( clock cycle : 0 ~ 15 )
//		RPO		: read pulse width ( clock cycle  : 0 ~ 15 )
//
// =============================================================================
// I/O PORTS	:
//	-- reg_i bit assignment --
//	input						 		clk;	(MSB)
//	input						 		rst_n;	  |
//	input								srst;	  |
//	input								enable;	  |
//	input	[ ADW - 1 : 0 ]				adr_i;	  |
//	input						 		wen_i;	  |
//	input								ren_i;	  |
//	input	[ ( DTW / 8 ) - 1 : 0 ]		ben_i;	  |
//	input	[ DTW - 1 : 0 ]				sig_i;	  |
//	input	[ DTW - 1 : 0 ]				wdata_i;(LSB)
//
// -- reg_o bit assignment --
//	output								wpls_o;	(MSB)
//	output								rpls_o;	  |
//	output	[ DTW - 1 : 0 ]				sg_o;	  |
//	output	[ DTW - 1 : 0 ]				rd_o;	(LSB)
//
// =============================================================================
`default_nettype none
`timescale 1ns / 100ps

module avb_frm_wt_reg_ff (
	reg_i	,
	reg_o
) ;

// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter ADW = 32;					// address bit width
	parameter DTW = 32;					// data bit width
	parameter ADR = 32'h0000_0000;		// address
	parameter INI = 32'h0000_0000;		// initial data
	parameter VLD = 32'hffff_ffff;		// generate register on/off every bit
	parameter PLS = 32'hffff_ffff;		// generate pulse output on/off every bit
	parameter TYP = 0;					// register type
										// 0: read/write register
										// 1: write-clear register
										// 2: read-only register
										// 3: read-clear register

	parameter WPO = 4'h0;				// write pulse width ( clock cycle : 0 ~ 15 )
	parameter RPO = 4'h0;				// read pulse width ( clock cycle  : 0 ~ 15 )

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	IN_PORT_WIDTH			= ADW + ( DTW / 8 ) + ( DTW * 2 ) + 6 ;
	parameter	OUT_PORT_WIDTH			= ( DTW * 2 ) + 2 ;
// =============================================================================
// PORT DECLARATION
// =============================================================================
	input	wire	[  IN_PORT_WIDTH-1 : 0 ]	reg_i	;
	output	wire	[ OUT_PORT_WIDTH-1 : 0 ]	reg_o	;

// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================
	wire								clk				;
	wire								rst_n			;
	wire								srst			;
	wire								enable			;
	wire	[		  ADW - 1 : 0 ]		in_address		;
	wire								in_write_en		;
	wire								in_read_en		;
	wire	[ ( DTW / 8 ) - 1 : 0 ]		in_byte_en		;
	wire	[		  DTW - 1 : 0 ]		in_signal		;
	wire	[		  DTW - 1 : 0 ]		in_wdata		;
	wire								out_wpulse		;
	wire								out_rpulse		;
	wire	[		  DTW - 1 : 0 ]		out_rdata		;
	wire	[		  DTW - 1 : 0 ]		out_signal		;

	wire	[		  DTW - 1 : 0 ]		in_bit_en		;
	wire								in_bit_en_flg	;

	reg		[		  DTW - 1 : 0 ]		in_signal_lat	;

	reg		[		  DTW - 1 : 0 ]		q_ff			;

	reg		[		  DTW - 1 : 0 ]		pls_ff			;

	reg		[				3 : 0 ]		w_pls_cnt_ff	;
	reg		[				3 : 0 ]		r_pls_cnt_ff	;
	reg									w_pls_ff		;
	reg									r_pls_ff		;


	genvar								i				;
	genvar								j				;

// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================
//===============
// common input
//===============
	assign	clk								=	reg_i [ ADW + ( DTW / 8 ) + ( DTW * 2 ) + 5 ] ;
	assign	rst_n							=	reg_i [ ADW + ( DTW / 8 ) + ( DTW * 2 ) + 4 ] ;
	assign	srst							=	reg_i [ ADW + ( DTW / 8 ) + ( DTW * 2 ) + 3 ] ;
	assign	enable							=	reg_i [ ADW + ( DTW / 8 ) + ( DTW * 2 ) + 2 ] ;
	assign	in_address						=	reg_i [ ADW + ( DTW / 8 ) + ( DTW * 2 ) + 1 : ( DTW / 8 ) + ( DTW * 2 ) + 2 ] ;
	assign	in_write_en						=	reg_i [ ( DTW / 8 ) + ( DTW * 2 ) + 1 ] ;
	assign	in_read_en						=	reg_i [ ( DTW / 8 ) + ( DTW * 2 ) ] ;
	assign	in_byte_en						=	reg_i [ ( DTW / 8 ) + ( DTW * 2 ) - 1 : ( DTW * 2 ) ] ;
	assign	in_signal						=	reg_i [ ( DTW * 2 ) - 1 : DTW ] ;
//=========
// output
//=========
	assign	reg_o [ ( DTW * 2 ) + 1 ]		=	out_wpulse ;
	assign	reg_o [ DTW * 2 ]				=	out_rpulse ;
	assign	reg_o [ ( DTW * 2 ) - 1 : DTW ]	=	out_signal ;
	assign	reg_o [ DTW - 1 : 0 ]			=	out_rdata ;


	generate

		for ( i = 0 ; i < DTW / 8 ; i = i + 1 ) begin : generate_bit_enable
			assign	in_bit_en [ ( ( i + 1 ) * 8 ) - 1 : ( i * 8 ) ] = { 8 { in_byte_en [ i ] } } ;
		end

	endgenerate

	assign	in_bit_en_flg					=	| in_bit_en ;


	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			in_signal_lat					<=	{ DTW { 1'b0 } } ;
		end
		else if ( enable ) begin
			in_signal_lat					<=	in_signal ;
		end
		else begin
			in_signal_lat					<=	in_signal_lat ;
		end
	end


//=================================
// register & out_signal generate
//=================================
	generate

		if ( TYP == 0 ) begin: read_write_register

			for ( j = 0 ; j < DTW ; j = j + 1 ) begin : generate_register

				if ( ( VLD [ j ] == 1'b1 ) && ( PLS [ j ] == 1'b1 ) ) begin

					assign	in_wdata [ j ]		=	reg_i [ j ] ;

					always @( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( srst ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] ) begin
								q_ff [ j ]		<=	in_wdata [ j ] ;
							end
						end
					end

					assign	out_rdata [ j ]		=	q_ff [ j ] ;

					always @ ( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( srst ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] && ( WPO != 4'h0 ) ) begin
								pls_ff [ j ]	<=	1'b1 ;
							end
							else if ( w_pls_ff ) begin
								pls_ff [ j ]	<=	pls_ff [ j ] ;
							end
							else begin
								pls_ff [ j ]	<=	1'b0 ;
							end
						end
					end

					assign	out_signal [ j ]	=	pls_ff [ j ] & w_pls_ff ;

				end
				else if ( ( VLD [ j ] == 1'b1 ) && ( PLS [ j ] == 1'b0 ) ) begin

					assign	in_wdata [ j ]		=	reg_i [ j ] ;
					
					always @( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( srst ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] ) begin
								q_ff [ j ]		<=	in_wdata [ j ] ;
							end
						end
					end

					assign	out_rdata [ j ]		=	q_ff [ j ] ;
					assign	out_signal [ j ]	=	1'b0 ;

				end
				else if ( ( VLD [ j ] == 1'b0 ) && ( PLS [ j ] == 1'b1 ) ) begin

					assign	out_rdata [ j ]		=	1'b0 ;

					always @ ( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( srst ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] && ( WPO != 4'h0 ) ) begin
								pls_ff [ j ]	<=	1'b1 ;
							end
							else if ( w_pls_ff ) begin
								pls_ff [ j ]	<=	pls_ff [ j ] ;
							end
							else begin
								pls_ff [ j ]	<=	1'b0 ;
							end
						end
					end

					assign	out_signal [ j ]	=	pls_ff [ j ] & w_pls_ff ;

				end
				else begin

					assign	out_rdata [ j ]		=	1'b0 ;
					assign	out_signal [ j ]	=	1'b0 ;

				end
			end

		end
		else if ( TYP == 1 ) begin: write_clear_register

			for ( j = 0 ; j < DTW ; j = j + 1 ) begin : generate_register

				if ( ( VLD [ j ] == 1'b1 ) && ( PLS [ j ] == 1'b1 ) ) begin

					assign	in_wdata [ j ]		=	reg_i [ j ] ;

					always @( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( srst ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] && in_wdata [ j ] ) begin
								q_ff [ j ]		<=	1'b0  ;
							end
							else if ( in_signal [ j ] && !( in_signal_lat [ j ] ) ) begin
								q_ff [ j ]		<= 	1'b1 ;
							end
						end
					end

					assign	out_rdata [ j ]		=	q_ff [ j ]	;

					always @ ( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( srst ) begin
							pls_ff [ j ]		<=	1'b0 ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] && ( WPO != 4'h0 ) ) begin
								pls_ff [ j ]	<=	1'b1 ;
							end
							else if ( w_pls_ff ) begin
								pls_ff [ j ]	<=	pls_ff [ j ] ;
							end
							else begin
								pls_ff [ j ]	<=	1'b0 ;
							end
						end
					end

					assign	out_signal [ j ]	=	pls_ff [ j ] & w_pls_ff ;

				end
				else if ( VLD [ j ] == 1'b1 ) begin

					assign	in_wdata [ j ]		=	reg_i [ j ] ;
					
					always @( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( srst ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_write_en && in_bit_en [ j ] && in_wdata [ j ] ) begin
								q_ff [ j ]		<=	1'b0  ;
							end
							else if ( ( in_signal [ j ] && !( in_signal_lat [ j ] ) ) ) begin
								q_ff [ j ]		<= 	1'b1 ;
							end
						end
					end

					assign	out_rdata [ j ]		=	q_ff [ j ] ;
					assign	out_signal [ j ]	=	1'b0 ;

				end
				else begin

					assign	out_rdata [ j ]		=	1'b0 ;
					assign	out_signal [ j ]	=	1'b0 ;

				end
			end

		end
		else if ( TYP == 2 ) begin: read_only_register

			for ( j = 0 ; j < DTW ; j = j + 1 ) begin : generate_register

				assign	out_signal [ j ]	=	1'b0 ;

			end

			assign out_rdata					=	in_signal ;


		end
		else if ( TYP == 3 ) begin: read_clear_register

			for ( j = 0 ; j < DTW ; j = j + 1 ) begin : generate_register

				if ( VLD [ j ] == 1'b1 ) begin
					
					always @( posedge clk or negedge rst_n ) begin
						if ( !rst_n ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( srst ) begin
							q_ff [ j ]			<=	INI [ j ] ;
						end
						else if ( enable ) begin
							if ( ( ADR == in_address ) && in_read_en ) begin
								q_ff [ j ]		<=	1'b0  ;
							end
							else if ( in_signal [ j ]  && !( in_signal_lat [ j ] ) ) begin
								q_ff [ j ]		<=	1'b1 ;
							end
						end
					end

					assign	out_rdata [ j ]		=	q_ff [ j ] ;
					assign	out_signal [ j ]	=	1'b0 ;

				end
				else begin

					assign	out_rdata [ j ]		=	1'b0 ;
					assign	out_signal [ j ]	=	1'b0 ;

				end

			end

		end
	endgenerate

//=======================
// write pulse generate
//=======================

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			w_pls_ff				<=	1'b0 ;
			w_pls_cnt_ff			<=	4'h0 ;
		end
		else if ( srst ) begin
			w_pls_ff				<=	1'b0 ;
			w_pls_cnt_ff			<=	4'h0 ;
		end
		else if ( enable ) begin
			if ( w_pls_cnt_ff == WPO ) begin
				w_pls_ff			<=	1'b0 ;
				w_pls_cnt_ff		<=	4'h0 ;
			end
			else if ( ( ADR == in_address ) && in_write_en && w_pls_ff && in_bit_en_flg ) begin
				w_pls_ff			<=	1'b1 ;
				w_pls_cnt_ff		<=	4'h1 ;
			end
			else if ( ( ADR == in_address ) && in_write_en && ( WPO != 4'h0 ) && in_bit_en_flg ) begin
				w_pls_ff			<=	1'b1 ;
				w_pls_cnt_ff		<=	w_pls_cnt_ff + 1'b1 ;
			end
			else if ( w_pls_ff ) begin
				w_pls_ff			<=	1'b1 ;
				w_pls_cnt_ff		<=	w_pls_cnt_ff + 1'b1 ;
			end
			else begin
				w_pls_ff			<=	1'b0 ;
				w_pls_cnt_ff		<=	4'h0 ;
			end
		end
	end

	assign out_wpulse				=	w_pls_ff ;

//=======================
// read pulse generate
//=======================
	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			r_pls_ff				<=	1'b0 ;
			r_pls_cnt_ff			<=	4'h0 ;
		end
		else if ( enable ) begin
			if ( r_pls_cnt_ff == RPO ) begin
				r_pls_ff			<=	1'b0 ;
				r_pls_cnt_ff		<=	4'h0 ;
			end
			else if ( ( ADR == in_address ) && in_read_en && r_pls_ff ) begin
				r_pls_ff			<=	1'b1 ;
				r_pls_cnt_ff		<=	4'h1 ;
			end
			else if ( ( ADR == in_address ) && in_read_en && ( RPO != 4'h0 ) ) begin
				r_pls_ff			<=	1'b1 ;
				r_pls_cnt_ff		<=	r_pls_cnt_ff + 1'b1 ;
			end
			else if ( r_pls_ff ) begin
				r_pls_ff			<=	1'b1 ;
				r_pls_cnt_ff		<=	r_pls_cnt_ff + 1'b1 ;
			end
			else begin
				r_pls_ff			<=	1'b0 ;
				r_pls_cnt_ff		<=	4'h0 ;
			end
		end
	end

	assign out_rpulse				=	r_pls_ff ;


endmodule

`default_nettype wire
