// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_divider.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Yuichi Shinoda
// DATE			: 08/22/2013
// =============================================================================
// DESCRIPTION	: divider circuit
//
// =============================================================================
// REVISION HISTORY	:
//	v1.0	: Initial version release
//
// =============================================================================
// PARAMETERS	:
//		DVEND 			: dividend bit width 2 ~ 32
//		DVSOR 			: divisor bit width 2 ~ 32
//		PIPELINE		: pipeline architecture 0:non pipeline / 1:pipeline
//
// =============================================================================
// I/O PORTS	:
//		clk				: clock for all circuit
//		rst_n			: asynchronous reset ( low active )
//		srst			: synchronous reset
//		enable			: clock enable
//	
//		valid_i			: data valid of input data
//		dividend_i		: dividend data
//		divisor_i		: divisor data
//	
//		ready_o			: ready signal
//		valid_o			: data valid of output data
//		quotient_o		: quotient data
//		remainder_o		: remainder data
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module	avb_pkt_asm_divider (
	clk				,
	rst_n			,
	srst			,
	enable			,

	valid_i			,
	dividend_i		,
	divisor_i		,

	ready_o			,
	valid_o			,
	quotient_o		,
	remainder_o
) ;
	
// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter DVEND 					= 8 ;			// dividend bit width 2 ~ 32
	parameter DVSOR 					= 8 ;			// divisor bit width 2 ~ 32
	parameter PIPELINE					= 1 ;			// pipeline architecture 0:non pipeline / 1:pipeline

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter INT_DVEND 				= DVEND + DVSOR ;
	parameter BUSY_TIMER				= 8 ;
	parameter BUSY_TIMER_OVF			= DVEND - 1 ;
	parameter PP_STEP					= DVEND ;
	
	parameter DVEND_INT_BIT				= INT_DVEND * PP_STEP ;
	parameter DVSOR_INT_BIT				= DVSOR * PP_STEP ;

	parameter DIV_IDLE					= 3'b001 ;
	parameter DIV_CALC_BUSY				= 3'b010 ;
	parameter DIV_RSLT_OUT				= 3'b100 ;
	
// =============================================================================
// PORT DECLARATION
// =============================================================================
	
	input	wire							clk			;	// clock for all circuit
	input	wire							rst_n		;	// asynchronous reset ( low active )
	input	wire							srst		;	// synchronous reset
	input	wire							enable		;	// clock enable

	input	wire							valid_i		;	// data valid of input data
	input	wire	[ DVEND-1 : 0 ] 		dividend_i	;	// dividend data
	input	wire	[ DVSOR-1 : 0 ] 		divisor_i	;	// divisor data

	output	wire							ready_o		;	// data reception ready of input data
	output	wire							valid_o		;	// data valid of output data
	output	wire	[ DVEND-1 : 0 ] 		quotient_o	;	// quotient data
	output	wire	[ DVSOR-1 : 0 ] 		remainder_o	;	// remainder data

// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================
	
	reg 	[				2 : 0 ]	div_state_ff	;
	reg 							busy_ff			;
	reg 	[	 BUSY_TIMER-1 : 0 ]	busy_cnt_ff		;
	reg 							valid_ff		;
	
	wire							enable_in		;
	wire	[		PP_STEP-1 : 0 ]	valid_int		;
	wire	[	  INT_DVEND-1 : 0 ] dividend_in		;
	wire	[ DVEND_INT_BIT-1 : 0 ]	dividend_int	;
	wire	[	  INT_DVEND-1 : 0 ] dividend_out	;
	wire	[		  DVSOR-1 : 0 ] divisor_in		;
	wire	[ DVSOR_INT_BIT-1 : 0 ]	divisor_int		;
	wire	[		  DVSOR-1 : 0 ] divisor_out		;
	
	genvar							i				;
	
// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================
	
	generate
		
		//-----------------------//
		///--  PIPELINE = YES  --//
		//-----------------------//
		if ( ( PIPELINE == 1 ) || ( PP_STEP == 1 ) ) begin	: if_label
			
			assign ready_o				= enable ;
			assign valid_o				= valid_int[PP_STEP-1] ;
			assign quotient_o 			= dividend_int[INT_DVEND*(PP_STEP-1)+DVEND-1:INT_DVEND*(PP_STEP-1)] ;
			assign remainder_o			= dividend_int[(INT_DVEND*PP_STEP)-1:(INT_DVEND*PP_STEP)-DVSOR] ;
			
			for ( i = 0 ; i < PP_STEP ; i = i + 1 ) begin		: pp_radix2
				
				if ( i == 0 ) begin 								: if_label
					avb_pkt_asm_div_radix2
						#(
							.DVEND		( DVEND	) ,
							.DVSOR		( DVSOR	)
						)
						u_div_radix2 (
							.clk		( clk											) ,
							.rst_n		( rst_n											) ,
							.srst		( srst											) ,
							.enable 	( enable										) ,
							
							.valid_i	( valid_i										) ,
							.dividend_i	( { {(DVSOR+DVEND-DVEND){1'b0}} , dividend_i }	) ,
							.divisor_i	( divisor_i										) ,
							
							.valid_o	( valid_int[i]									) ,
							.dividend_o	( dividend_int[(i+1)*INT_DVEND-1:i*INT_DVEND]	) ,
							.divisor_o	( divisor_int[(i+1)*DVSOR-1:i*DVSOR]			)
						) ;
				end
				else begin											: else_label
					avb_pkt_asm_div_radix2
						#(
							.DVEND				( DVEND	) ,
							.DVSOR				( DVSOR	)
						)
					u_div_radix2 (
						.clk			( clk											) ,
						.rst_n			( rst_n											) ,
						.srst			( srst											) ,
						.enable 		( enable										) ,
						
						.valid_i		( valid_int[i-1]								) ,
						.dividend_i		( dividend_int[i*INT_DVEND-1:(i-1)*INT_DVEND]	) ,
						.divisor_i	 	( divisor_int[i*DVSOR-1:(i-1)*DVSOR]			) ,
						
						.valid_o		( valid_int[i]									) ,
						.dividend_o		( dividend_int[(i+1)*INT_DVEND-1:i*INT_DVEND]	) ,
						.divisor_o		( divisor_int[(i+1)*DVSOR-1:i*DVSOR]			)
					) ;
				end
				
			end
			
		end
		//---------------------//
		//--  PIPELINE = NO  --//
		//---------------------//
		else begin											: else_label
			
			assign ready_o				= ( div_state_ff == DIV_IDLE )
										| ( ( div_state_ff == DIV_RSLT_OUT ) & enable ) ;
			assign valid_o				= valid_ff ;
			assign quotient_o 			= dividend_out[DVEND-1:0] ;
			assign remainder_o			= dividend_out[INT_DVEND-1:DVEND] ;
			assign enable_in			= ( ( div_state_ff == DIV_IDLE ) & valid_i )
										| ( div_state_ff == DIV_CALC_BUSY )
										| ( ( div_state_ff == DIV_RSLT_OUT ) & valid_i & enable ) ;
			assign dividend_in			= ( busy_ff )? dividend_out : { {(DVSOR+DVEND-DVEND){1'b0}} , dividend_i } ;
			assign divisor_in			= ( busy_ff )? divisor_out : divisor_i ;
			
			avb_pkt_asm_div_radix2
				#(
					.DVEND				( DVEND	) ,
					.DVSOR				( DVSOR	)
				)
				uDIV_RADIX2 (
					.clk			( clk			) ,
					.rst_n			( rst_n			) ,
					.srst			( srst			) ,
					.enable 		( enable_in		) ,
					
					.valid_i		( 1'b0			) ,
					.dividend_i		( dividend_in	) ,
					.divisor_i	 	( divisor_in	) ,
					
					.valid_o		( 				) ,
					.dividend_o		( dividend_out	) ,
					.divisor_o		( divisor_out	)
				) ;
			
		end
		
	endgenerate
	
	generate
		if ( ( PIPELINE != 1 ) && ( PP_STEP != 1 ) ) begin	: if_label2
			
			always @( posedge clk or negedge rst_n ) begin
				if ( !rst_n ) begin
					div_state_ff	<= DIV_IDLE ;
					busy_ff 		<= 1'b0 ;
					busy_cnt_ff 	<= {BUSY_TIMER{1'b0}} ;
					valid_ff		<= 1'b0 ;
				end
				else if ( srst ) begin
					div_state_ff	<= DIV_IDLE ;
					busy_ff 		<= 1'b0 ;
					busy_cnt_ff 	<= {BUSY_TIMER{1'b0}} ;
					valid_ff		<= 1'b0 ;
				end
				else begin
					case ( div_state_ff )
						DIV_IDLE		: begin
							if ( valid_i ) begin
								div_state_ff	<= DIV_CALC_BUSY ;
								busy_ff 		<= 1'b1 ;
								busy_cnt_ff 	<= busy_cnt_ff + 1'b1 ;
								valid_ff		<= 1'b0 ;
							end
							else begin
								busy_ff 		<= 1'b0 ;
								busy_cnt_ff 	<= {BUSY_TIMER{1'b0}} ;
								valid_ff		<= 1'b0 ;
							end
						end
						DIV_CALC_BUSY	: begin
							if ( busy_cnt_ff == BUSY_TIMER_OVF ) begin
								div_state_ff	<= DIV_RSLT_OUT ;
								busy_ff 		<= 1'b0 ;
								busy_cnt_ff 	<= {BUSY_TIMER{1'b0}} ;
								valid_ff		<= 1'b1 ;
							end
							else begin
								busy_cnt_ff 	<= busy_cnt_ff + 1'b1 ;
							end
						end
						DIV_RSLT_OUT	: begin
							if ( valid_i & enable ) begin
								div_state_ff	<= DIV_CALC_BUSY ;
								busy_ff 		<= 1'b1 ;
								busy_cnt_ff 	<= busy_cnt_ff + 1'b1 ;
								valid_ff		<= 1'b0 ;
							end
							else if ( enable ) begin
								div_state_ff	<= DIV_IDLE ;
								valid_ff		<= 1'b0 ;
							end
						end
						default 	: begin
							div_state_ff	<= DIV_IDLE ;
							busy_ff 		<= 1'b0 ;
							busy_cnt_ff 	<= {BUSY_TIMER{1'b0}} ;
							valid_ff		<= 1'b0 ;
						end
					endcase
				end
			end
			
		end
	endgenerate
	
endmodule

`default_nettype wire
