// =============================================================================
// Copyright (C) MACNICA inc.
// =============================================================================
// FILE NAME	: avb_pkt_asm_to32.v
// DEPARTMENT	: MACNICA Strategic Technology Group
// AUTHOR		: Ryosuke Yamamoto
// DATE 		: 03/05/2014
// =============================================================================
// DESCRIPTION	:
//		
// =============================================================================
// REVISION HISTORY :
//		v1.0		: Initial Release
//
// =============================================================================
// PARAMETERS	:
//		Q_BIT		: Quantized bit width per color plane
//		PLANE		: Color plane count
//		SIZE_BIT	: Size setting signal bit width
//
// =============================================================================
// I/O PORTS	:
//		clk					: Clock for all circuit
//		rst_n				: Asynchronous reset ( low active )
//		srst				: Synchronous reset
//
//		valid_i				: Data valid of input VSI frame data.
//		frame_start_i		: Frame start of input VSI frame data.
//		frame_end_i			: Frame end of input VSI frame data.
//		field_i				: Field status of input VSI frame data.
//							  0:Odd field  1:Even field
//		pixel_i				: Pixel data of input VSI frame data.
//		width_i				: Horizontal size of input VSI frame data
//		height_i			: Vertical size of input frame data
//		ready_o				: Data reception ready of input VSI frame data.
//
//		valid_o				: Data valid of output VSI frame data.
//		frame_start_o		: Frame start of output VSI frame data.
//		frame_end_o			: Frame end of output VSI frame data.
//		field_o				: Field status of output VSI frame data.
//							  0:Odd field  1:Even field
//		pixel_o				: Pixel data of output VSI frame data.
//		width_o				: Horizontal size of output VSI frame data
//		height_o			: Vertical size of output frame data
//		ready_i				: Data reception ready of output VSI frame data.
//
//
// =============================================================================
`timescale 1ps/1ps
`default_nettype none

module avb_pkt_asm_to32 (
	clk					,
	rst_n				,
	srst				,

	valid_i				,
	frame_start_i		,
	frame_end_i			,
	field_i				,
	data_i				,
	width_i				,
	height_i			,
	ready_o				,

	valid_o				,
	frame_start_o		,
	frame_end_o			,
	field_o				,
	data_o				,
	width_o				,
	height_o			,
	ready_i				

) ;


// =============================================================================
// PARAMETER DEFINITION
// =============================================================================

	// ---------------------------------------------------------------------
	// Below parameters have to be defined from upper module
	// ---------------------------------------------------------------------
	parameter	SIZE_BIT			= 11 ;

	// ---------------------------------------------------------------------
	// Please do not change the following parameters
	// ---------------------------------------------------------------------
	parameter	IN_DATA_BIT			= 16 ;
	parameter	OUT_DATA_BIT		= 32 ;
	parameter	PKT_SIZE			= 11 ;


// =============================================================================
// PORT DECLARATION
// =============================================================================

	input	wire							clk					;
	input	wire							rst_n				;
	input	wire							srst				;

	input	wire							valid_i				;
	input	wire							frame_start_i		;
	input	wire							frame_end_i			;
	input	wire							field_i				;
	input	wire	[ IN_DATA_BIT-1 : 0 ]	data_i				;
	input	wire	[  SIZE_BIT-1 : 0 ]		width_i				;
	input	wire	[  SIZE_BIT-1 : 0 ]		height_i			;
	output	wire							ready_o				;

	output	wire							valid_o				;
	output	wire							frame_start_o		;
	output	wire							frame_end_o			;
	output	wire							field_o				;
	output	wire	[ OUT_DATA_BIT-1 : 0 ]	data_o				;
	output	wire	[  SIZE_BIT-1 : 0 ]		width_o				;
	output	wire	[  SIZE_BIT-1 : 0 ]		height_o			;
	input	wire							ready_i				;


// =============================================================================
// REG / WIRE DECLARATION
// =============================================================================

	wire							frame_start		;
	wire							frame_end		;

	reg		[ IN_DATA_BIT-1 : 0 ]	data_lat_ff		;
	reg								valid_mask_ff	;

// =============================================================================
// FUNCTION DESCRIPTION
// =============================================================================

	assign	frame_start	= valid_i & frame_start_i & ~( |data_i[3:0] ) ;
	assign	frame_end	= valid_i & frame_end_i ;

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			data_lat_ff	<= { IN_DATA_BIT{1'b0} } ;
		end
		else if ( srst ) begin
			data_lat_ff	<= { IN_DATA_BIT{1'b0} } ;
		end
		else if ( ready_i && valid_i ) begin
			data_lat_ff	<= data_i ;
		end
	end

	always @ ( posedge clk or negedge rst_n ) begin
		if ( !rst_n ) begin
			valid_mask_ff	<= 1'b0 ;
		end
		else if ( srst ) begin
			valid_mask_ff	<= 1'b0 ;
		end
		else if ( ready_i ) begin
			if ( frame_start ) begin
				valid_mask_ff	<= 1'b1 ;
			end
			else if ( frame_end ) begin
				valid_mask_ff	<= 1'b0 ;
			end
			else if ( valid_i ) begin
				valid_mask_ff	<= ~valid_mask_ff ;
			end
		end
	end

	assign	ready_o			= ready_i ;
	assign	valid_o			= ( ~valid_mask_ff & valid_i) | frame_end ;
	assign	frame_start_o	= frame_start_i ;
	assign	frame_end_o		= frame_end_i ;
	assign	field_o			= field_i ;
	assign	data_o			= { data_lat_ff , data_i } ;
	assign	width_o			= width_i ;
	assign	height_o		= height_i ;


endmodule
`default_nettype wire